; PROCTYPE_X1_ASM
; on H38
; 13th March 2000
;
;	  p r o c t y p e
;
;  This routine is used to detect the processor type.
;
;  Support routine used internally by various of the C68 startup modules.
;  This routine MUST be called in Supervisor mode, and with the
;  address of the system variables in a6.
;
;  Strictly speaking we may not need to do this for the processor type,
;  as there is now a system variable that should already have been set
;  up to handle this.   If this has been set non-zero, then we will skip
;  over the processor detection code.
;
;  As a slight optimisation, we do not try to make any further checks
;  for hardware FPU if this system has a chip that does not use a VBR
;  to point to exception vectors (i.e. a 68008 or 68000).
;
;  If the processor is determined to be a 68020 or better, then a check
;  is made for the string identifying a Miracle Super Gold Card as then
;  no further checks must be made due to hardware deficiencies of the SGC.
;
;  When the processor type is determined, the results are stored
;  into the system variables at offset $a1.
;
;  The Processor type value is encoded as a byte field and put into the
;  system variable at $a1(SV.BASE).  This byte is encoded as follows:
;
;  Bits 4-7    0 = 68008 or 68000
;	     1 = 68010
;	     2 = 68020
;	     3 = 68030
;	     4 = 68040
;	     6 = 68060
;
;  It has been found that a number of programmers have assumed that bits
;  0-3 are zero, so the information about the MMU is stored at
;  $D1(SV.BASE) and that about the FPU is now stored in
;  the new system variable at $d0(SV.BASE).  These bytes are encoded as
;  follows:
;
;	 FPU				     MMU
;       -1   No FPU present			   No MMU present
;       1    68881 (or equivalent) found	   68851
;       2    68882 (or equivalent) found	     -
;       3      -				   68030
;       4    68040				   68040 or 68LC040
;       5    68040 + FPSP* already loaded	     -
;       6    68060				   68060 or 68LC060
;       7    68060 + FPSP* already loaded	     -
;       9    68881 + FPSP (Library version) loaded     -
;      10    68882	  "         "		     -
;      12    68040	  "         "		     -
;      14    68060	  "         "		     -
;
;  This program will always load an FPSP and this is accessible via
;  the routine at $D8(SV.BASE). A previously loaded FPSP* is presumably
;  a kernel version which emulates the unimplemented instructions via
;  the addresses in the VBR. Unfortunately FPSAVE alters this and supplies new
;  addresses thus imobilising the old FPSP*.
;
;
;   Register usage from this point, up to the relocation code are:
;       D7      Results of detection code so far
;       D4      Address of old VBR
;       A4      Address of new VBR (and therefore memory allocated)
;       A5      Saved A7 supervisor mode (to restore after exception vectors)
;       A6      Address of System Variables  (preserved)
;
;  AMENDMENT HISTORY
;  ~~~~~~~~~~~~~~~~~
;  10 Nov 93 DJW  - Made improvements to code for recognising chip details
;
;  12 Mar 94 DJW  - Changed start-up processor detection code to allow for
;		the Super Gold card having its vectors in ROM.
;
;  19 Mar 94 DJW  - Further changes to processor detection code (using ideas
;		contributed by Andreas Rudolf).  This new version is more
;		forward compatible as whenever possible it relocates the
;		exception vectors via the VBR to ensure they are in RAM.
;		Also, the information is now encoded according to the
;		QDOS standard for the system variable (at offset $A1) so
;		that it is no longer necessary to store it internally.
;
;  16 Apr 94 DJW  - Changes to processor detection code to allow for fact that
;		some systems (Super Gold Card in particular) can take the
;		Co-Processor Protocol Violation exception vector when it
;		had been assumed that the Line-F one would be taken.  This
;		does NOT, however, fix the screen corruption that can
;		occur at the same time.
;
;  10 Oct 94 DJW  - The code for detecting processor code put into its own
;		file as part of the re-organisation of the C68 start-up
;		code and the commoning up of code with the RLM.
;	        - Comments from original crt_x file that related to
;		processor detection retained in this file.
;
;  20 Sep 95 SNG  - Faster test for 68020+
;	        - Numerous typos in comments fixed
;
;   6 Nov 96 GDG  - Setting byte $D0 of system variables
;	        - Not setting MMU code in system variables
;	        - Distinguishing between 68881 and 68882
;	        - No attempt made to see if FPSP present
;	        - Distinguishes between 68040 and 68060
;	        - Code altered for assembly by GWASS
;
;  10 Dec 96 GDG  - Setting the MMU code at $D1(SYS.BASE)
;
;  13 Mar 97 GDG  - Correcting code to distinguish between 68040 and 68060
;		and to test for MMU for 68040/60
;	        - Disarming FPU trap if set
;	        - Detecting whether if $D0(SYS.BASE) or $D1(SYS.BASE) is
;		set it is correct.
;
; 13 Mar 00 GDG   - Leaving $A1(SYS.BASE) alone if already set.
;	        - Setting $D0(SYS.BASE) correctly.
;
;
;
; BUGS REMAINING: (20 Sep 95)
;
; - Thinks the 68881 is a 68882 if the FPU is busy when this code runs.
; - No way to distinguish basic 68040/68060s from fully extended FPSPs.
;   This means code cannot safely rely on the full 68882 instruction set.
;   I suggest using bits 1..3 to flag FPU types including 68040 and 68060
;   with or without FPSP emulation of 68881/68882 opcodes = 7 FP variants.
; - Thinks 68060 is a 68040; should check PCR and tweak flags accordingly.
;
;----------------------------------------------------------------------------

;   .text

FPU	EQU   2		      ; => hardware FPU (for D7)
MMU	equ   $01 	      ; => 68851 (or equivalent) MMU

IS881	equ   $04 	      ; => 68881 (or equivalent) FPU
IS882	equ   $08 	      ; => 68882 (or equivalent) FPU

IS000	equ   $00 	      ; => 68000
IS010	equ   $10 	      ; => 68010
IS020	equ   $20 	      ; => 68020
IS030	equ   $30 	      ; => 68030
IS040	equ   $40 	      ; => 68040
IS060	equ   $60 	      ; => 68060

ILLVEC	equ   $10 	      ; Illegal instruction vector
LINEF	equ   $2c 	      ; Line F vector
COPROC	equ   $34 	      ; CoProcessor Protocol Violation
DZ_VEC	equ   $C8 	      ; Divide by Zero FP vector

; MEM.ACHP equ    $c0	    ; Allocate Common Heap vector
; MEM.RCHP equ    $c2	    ; Release Common heap vector

PROCTYPE:
;    movem.l d3/a2-a5,-(sp)	   ; save registers corrupted
	MOVEM.L	D1-D5/A0-5,-(A7)
    move.l  a7,a5		  ; save a7 for convenience later

;   If the processor type information is already known, then
;   there is not much point in trying to determine it again.
;   We can therefore skip all this magic mumbo jumbo
;   (Except that we need the FPU information)

	MOVEQ	#0,D7
	MOVE.B	$A1(A6),D7   ; System Variable -> D7
	ANDI.W	#$F0,D7	   ; Clearing MMU bits - added 13 Mar 00
;    tst.b   $a1(a6)	   ; System Variable set ?
;    bne	   PROCTYPE_END	   ; ... YES, then details already known
	BNE	HALF_DONE    ; ... YES, but FPU/MMU state to be found

;   Check for 68020 or better.
;   We did this by realizing that the processors before the
;   68020 ignored any scaling bits in an instruction.
;   Updated by SNG to use faster and shorter MOVEM technique.
;   (But this failed to set D7 for 68000/8/10)
;
;   moveq   #1,d0		  ; Set index register
;   lea	  cputable(pc),a0	  ; A0 -> table
;   dc.l	  $1e300C00
;   MOVE.B  0(A0,D0.L*4),D7	  ; get return value (0 for 68008/68000/68010)

START1:
    movem.l a7,-(a7)	  ; See if the pre-decrement is pipelined?
    cmp.l   (a7)+,a7
    bne	  GOT_020 	  ; It is an 020 or better

;   We only have a 68008, 68000 or 68010
;   if we get to this point.	We treat the
;   68008 and 68000 both as if they were the
;   68000 as they are completely code compatible

    lea	  ILLVEC,a2	  ; set A2 to point to vector we will change
    move.l  (a2),d5 	  ; save old vector value
    lea	  GOT_000(pc),a0	  ; A0 -> Trap handler
    move.l  a0,(a2) 	  ; Set new vector
    cmpa.l  (a2),a0 	  ; did we change it ?
    bne	  END_CLEAR	  ; ... NO, then it is ROM - assume only 68000
;			  ; we cannot check for MMU and FPU,
;    dc.w    $42c0
    MOVE	  CCR,D0		  ; see if 010 (illegal on 68008/68000)

;   If we got here it is a 68010

    moveq   #IS010,d7	  ; it is an 010
GOT_000:
    move.l  a5,a7		  ; restore stack from a6
    move.l  d5,(a2) 	  ; restore vector
    tst.l   d7		  ; was it a 68010 ?
    beq	  END_CLEAR	  ; ... NO, then we are finished
GOT_010:
    bsr	  SETUP_NEW_VBR	  ; set VBR etc
    bra	  FPUTYPE 	  ; ... go for FPU type

HALF_DONE:
	MOVE.B	D7,D0
	LSR.B	#3,D0
	ANDI.W	#14,D0	  ; D0 = 2, 4, 6 ... 14
	LEA	JT,A0	  ; Load Jump Table
	ADDA.W	-2(A0,D0.W),A0
	JMP	(A0)	  ; We know the processor - so go directly
;
JT	DC.W	GOT_010-JT
	DC.W	GOT_20A-JT
	DC.W	GOT_30A-JT
	DC.W	GOT_40A-JT
	DC.W	NOTHING-JT  ; There should not be $50!!
	DC.W	GOT_40A-JT  ; 68040 and 68060 got to the same place
	DC.W	NOTHING-JT  ; There should not be $70!!
;
GOT_30A	ORI.B	#MMU,D7
	BRA	GOT_010
GOT_40A	BSR	SETUP_NEW_VBR
	BRA	GOT_40
NOTHING	MOVEQ	#0,D7
	BRA	START1
GOT_20A	BSET	#8,D7	  ; Mark truly 68020
	BRA	GOT_020B

;   Table used in distinguishing between processors below
;   68020 and those that are 68020 or better. Redundant - SNG
;
; cputable:   ; CPU value table
;   dc.l	    0
;   dc.b	    IS020,0,0,0
;
;   This routine is used to set up a new VBR.  It does the following tasks:
;       a)  Allocate memory for the new VBR
;       b)  Save the old VBR value into D4
;       c)  Copy the old vectors into the allocated memory
;       d)  Set the VBR to point to this value

;   It allocates memory, and makes a copy of the current vector table
;   to the address in a4.  The VBR is then changed to use this
;   new table.
;   (d1, a0 and a1 are corrupted)

SETUP_NEW_VBR:
;    dc.l    $4e7a4801
    MOVEC   VBR,D4		  ; current vector base register -> d4

;   Allocate memory to hold vector tables

    move.w  MM_ALCHP,a0	  ; Allocate common heap memory vector
    moveq   #65,d1		  ; We want 1024+16 - get by 65 * 16 which
    lsl.l   #6,d1		  ; saves 2 bytes over simple move of constant
    jsr	  (a0)		  ; do it
    lea	  16(a0),a4	  ; Go past header and store in a4

    move.w  #$FF,d1 	  ; count of entries in VBR - 1
    move.l  d4,a0		  ; old VBR address as source
    move.l  a4,a1		  ; new VBR address as target
COPY_LOOP:
    move.l  (a0)+,(a1)+	  ; move one entry
    dbra	  d1,COPY_LOOP	  ; ... go for next
    MOVEQ #7,D1		  ; to count 8 vectors
    LEA	  $C0(A4),A1	  ; 1st address (vector 48)
    LEA	FRESET,A0
COPY_LOOP1:
    MOVE.L  A0,(A1)+	  ; to arrange to reset the FPU
    DBF	  D1,COPY_LOOP1	  ; if an exception is pending
    MOVE.L  A0,$F4(A4)
;    dc.l    $4e7BC801
    MOVEC   A4,VBR		  ; a4 -> vector base register
    rts
;
FRESET:
    FRESTORE NULL
    RTE
;
NULL DC.L    0,0,0


;   If we get here, then we have at least a 68020
;   Set up LINE-F vector for 68020 or better CPU

GOT_020:

;   These next few lines are to handle problems with the Super Gold Card
;   on which issuing the LINE-F instructions can have undesirable side
;   effects due to hardware limitations.   Future releases of the Super
;   Gold Card will set the system variable so that this point will not
;   be reached, but this caters for all those already shipped.

	MOVEQ	#IS020,D7
GOT_020B:
    cmpi.l  #"Gold",$1000a	 ; Check for 'Gold' at this address
    beq	  GOT_SGC

    bsr	  SETUP_NEW_VBR	 ; Go to set up the VBR
	BCLR	#8,D7	 ; Was it a 68020 in Sys Vars?
	BNE	NOT_040	 ; ... YES

;   For a 68040/60, we need to invalidate the caches.  As a
;   by-product if this works we know that we have a 68040/60!

    lea	  NOT_040(pc),a0
    bsr	  SET_VECTORS
;    dc.w    $f4f8
    CPUSHA  BC		  ;  68040: flush and invalidate caches
    moveQ   #IS040,d7	  ;  ... it is an 68040/60

;
; Now see if is a 68060
;
	LEA	GOT_40,A0
	BSR	SET_VECTORS1 ; Illegal vector set if 68040
	MOVEC	PCR,D0
;
; It is a 68060
;
	MOVEQ	#IS060,D7

*   Now we can find out if the 68040/60 has a MMU
*   (i.e. it is a full 680x0 or a 68LC0x0 rather than a 68EC0x0)

GOT_40:
    lea	  FPUTYPE(pc),a0	  ; set vector again
    bsr	  SET_VECTORS1
;    dc.w    $f518
;   PFLUSHA
	MOVEC	TC,D0
    ORI.b   #MMU,d7
    bra	  FPUTYPE

;
;   This routine does what SET_VECTORS c) does and:
;       a)  Stores the address in A0 in the Illegal exception vector

SET_VECTORS1:
	MOVE.L	A0,ILLVEC(A4)
	BRA	SET_VECTORS2
;
;   This routine does the following:
;       a)  Stores the address in a0 in the LINE-F exception vector
;       b)  Stores the address in a0 in the Co-Processor violation vector
;	  This one is needed for systems (such as Super Gold Card) which
;	  take this vector rather than the LINE-F one (due to hardware
;	  limitations) when they do not recognise LINE-F instructions
;       c)  Tidies up the supervisor stack if there is any information
;	  left from earlier exceptions.

SET_VECTORS:
    move.l  a0,LINEF(a4)	  ; Set LINE-F vector
    move.l  a0,COPROC(a4)	  ; Set Co-Processor vector
SET_VECTORS2:
    move.l  (a7)+,a0	  ; get return address
    move.l  a5,a7		  ; ensure stack is tidy
    jmp	  (a0)		  ; return

;   We next decide if it is a 68030 or 68020+68551
;   If not, then the 68020/68040 is already set up.

NOT_040:
    lea	  FPUTYPE(pc),a0	  ; set vector
    bsr	  SET_VECTORS
;    dc.l    $f0002400
	68030		  ; Get the 68030/68851 version of PFLUSHA
    PFLUSHA		  ; flush MMU - 68030 or 68020+68851
	68040

;   If we got here we definitely have either
;   a 68030, or a 68020+68851.
;   In both cases we have a MMU, so flag that anyway

    ori.b   #MMU,d7 	  ; set MMU flag
    lea	  GOT_030(pc),a0	  ; set another vector
    bsr	  SET_VECTORS
;    dc.w    0xf127
    PSAVE   -(a7)		  ; only for 68851
    bra	  FPUTYPE 	  ; go and look for hardware FPU

GOT_030:
    ori.b   #IS030,d7	  ; we are an 68030

;   We next decide whether hardware FP is present

FPUTYPE:
    lea	  GOT_ALL,a0	  ; a0 -> trap handler
    bsr	  SET_VECTORS
;    dc.l    $f2800000
    FNOP			  ; See if we have hardware FP at all ?
	ORI.B	#FPU,D7	  ; Mark FPU present
	CMPI.B	#IS040,D7
	BGE	GOT_ALL	  ; 68040/60

;   We have hardware FP - is it 68881 or 68882

;    dc.w    $f327
FPU1	FSAVE	-(A7)
	TST.B	(A7)	  ; NULL?
	BNE	FPU2	  ; ... NO
	MOVEA.L	A5,A7	  ; reset stack
	FMOVE.B	#1,FP0	  ; Set not NULL
	BSET	#7,D7	  ; Mark 'was NULL'
	BRA	FPU1	  ; Go again
;
; The frames are either IDLE or BUSY
; The sizes are:
;
;	  68881	   68882
; IDLE	    $1C	     $3C
; BUSY	    $B8	     $D8
;
FPU2	MOVE.L	A5,D0
	SUB.L	A7,D0	  ; Size of frame
	CMPI.W	#$1C,D0
	BEQ	F81	  ; 68881
	CMPI.W	#$B8,D0
	BEQ	F81	  ; 68881
	ORI.B	#8,D7	  ; Mark 68882
	BRA	FPU3
F81	ORI.B	#4,D7	  ; Mark 68881
FPU3	BCLR	#7,D7	  ; If the FPU state was . . .
	BEQ	GOT_ALL	  ;
	CLR.L	-(A7)	  ; . . . NULL reset it
	FRESTORE	(A7)+

;   It is now time to tidy up as we have our answers

GOT_ALL:
    move.l  a5,a7		  ; restore stack
;    dc.l    $4e7b4801
    MOVEC   D4,VBR		  ; reset D4 -> vector base register

    lea	  -16(a4),a0	  ; get area to release (adjusted for header)
    moveA.w MM_RECHP,a1	  ; set vector required
    jsr	  (a1)		  ; do it

GOT_SGC:
END_CLEAR:
;
; We now set D0 = byte at $D0 of System Variables
;
	MOVEQ	#0,D0	  ; Initialise D0
	BCLR	#1,D7	  ; Do we have FPU? . .
	BNE	END_CL1	  ; . . YES
	MOVEQ	#-1,D0	  ; Mark 'no FPU'
	BRA	END_CL5
END_CL1	BCLR	#2,D7	  ; 68881? . . .
	BEQ	END_CL3	  ; . . . NO
	ORI.B	#1,D0	  ; Mark 68881
END_CL3	BCLR	#3,D7	  ; 68882? . . .
	BEQ	END_CL4	  ; . . . NO
	ORI.B	#2,D0	  ; Mark 68882
END_CL4	MOVE.B	D7,D1
	LSR.B	#4,D1
	CMPI.B	#4,D1	  ; 68040/60? . . .
	BLT	END_CL5	  ; . . . NO
	OR.B	D1,D0
END_CL5	TST.B	$D0(A6)
	BEQ	END_CL6	  ; SYS_FPU not yet set
	CMP.B	$D0(A6),D0  ; is SYS_FPU correctly set? . . .
	BNE	PROC_ERR	  ; . . . no
END_CL6	BSET	#3,D0	  ; mark FPSP
	SWAP	D7
	MOVE.W	D0,D7	  ; Set top word . . .
	SWAP	D7	  ; . . . to $D0(A6) code
	MOVE.B	D0,$D0(A6)
;
;    addq.b  #IS881,d7	   ; Assume 68881
;    cmp.b   #0x18,1(a7)	   ; is it a 68881?
;    beq	   GOT_ALL	   ; YES .. we are done
;    addq.b  #IS882-IS881,d7	   ; NO, make a 68882

	BCLR	#8,D7	  ; take off 'true 68020' marker
	MOVE.B	D7,D0
	MOVEQ	#-1,D1	  ; provisionally no MMU
	BCLR	#0,D0	  ; clear MMU bit (but leave in D7)
	BEQ	END_CL2	  ; no MMU
	BFEXTU	D7{25:3},D1
	CMPI.B	#2,D1
	BGT	END_CL2	  ; 68030+
	MOVEQ	#1,D1	  ; mark 68851
END_CL2	TST.B	$D1(A6)
	BEQ	END_CL7	  ; SYS_MMU not yet set
	CMP.B	$D1(A6),D1  ; is SYS_MMU correct? . . .
	BNE	PROC_ERR	  ; . . . no
END_CL7	MOVE.B	D1,$D1(A6)
;
	TST.B	$A1(A6)	  ; Is it already set? . .
	BNE	END_CL8	  ; . . yes
    move.b  d0,$a1(a6)	  ; save answers in system variables area
END_CL8	MOVEQ	#0,D0	  ; mark error free

PROCTYPE_END:
    movem.l (A7)+,d1-5/a0-a5	  ; save registers corrupted
    rts
;
PROC_ERR	MOVEQ	#-19,D0
	BRA	PROCTYPE_END
